# Much of this is based on lucidrains implementation of rotary embeddings
# https://github.com/lucidrains/rotary-embedding-torch

# Removed many of the configurable options to simplify the code
# and allow it to work with torch.compile
# Uses xpos

import torch
from torch import nn


class RotaryEmbedding(nn.Module):
    def __init__(
        self,
        dim,
        theta=10000,
        scale_base=512,
    ):
        super().__init__()

        freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
        self.register_buffer("freqs", freqs.unsqueeze(0))

        scale = (torch.arange(0, dim, 2) + 0.4 * dim) / (1.4 * dim)
        self.register_buffer("scale", scale)

        self.scale_base = scale_base

    def forward(self, q, k):
        seq_len = q.shape[-2]

        seq = torch.arange(seq_len, device=q.device, dtype=q.dtype)

        freqs = seq.unsqueeze(1) @ self.freqs.type_as(q)
        freqs = torch.repeat_interleave(freqs, 2, dim=-1)

        power = (seq - seq_len // 2) / self.scale_base
        scale = self.scale.pow(power.unsqueeze(-1)).type_as(q)

        scale = torch.cat((scale, scale), dim=-1)

        fc = freqs.cos()
        fs = freqs.sin()

        rotated_q = self.apply_rotary_emb(fc, fs, q, scale=scale)
        # Idk why but reciprocal will sometimes promote f16 to f32
        rotated_k = self.apply_rotary_emb(fc, fs, k, scale=scale.reciprocal().type_as(q))

        return rotated_q, rotated_k

    def apply_rotary_emb(self, freqs_cos, freqs_sin, t, scale):
        t = (t * freqs_cos * scale) + (self.rotate_half(t) * freqs_sin * scale)
        return t

    def rotate_half(self, x):
        d = x.size(-1) // 2
        x = x.view(*x.shape[:-1], d, 2)

        x1, x2 = x.unbind(dim=-1)

        x = torch.stack((-x2, x1), dim=-1)
        x = x.view(*x.shape[:-2], -1)
        return x



import copy
from math import log
from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from hnn_utils.nn.RotaryEmbedding import RotaryEmbedding


class RotaryEncoder(nn.Module):
    """
    PyTorch API compatible (more or less) encoder for Transformers, with support for rotary embeddings.

    Args:
        encoder_layer: encoder layer to use
        num_layers: number of encoder layers
        fix_init: Torch clones the encoder layer causing all layers to have identical initialization. This corrects that.
    """

    def __init__(
        self,
        encoder_layer: nn.Module,
        num_layers: int,
        fix_init: bool = True,
    ) -> None:
        super().__init__()

        self.layers = _get_clones(encoder_layer, num_layers)

        if fix_init:
            _transformer_reset(self)

    def forward(
        self,
        src: Tensor,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
    ) -> Tensor:
        output = src

        for layer in self.layers:
            output = layer(output, mask, src_key_padding_mask, is_causal=is_causal)

        return output


class RotaryDecoder(nn.Module):
    """
    PyTorch API compatible (more or less) decoder for Transformers, with support for rotary embeddings.

    Args:
        decoder_layer: decoder layer to use
        num_layers: number of decoder layers
        fix_init: Torch clones the encoder layer causing all layers to have identical initialization. This corrects that.
    """

    def __init__(
        self,
        decoder_layer: nn.Module,
        num_layers: int,
        fix_init: bool = True,
    ) -> None:
        super().__init__()

        self.layers = _get_clones(decoder_layer, num_layers)

        if fix_init:
            _transformer_reset(self)

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
    ) -> Tensor:
        output = tgt

        for layer in self.layers:
            output = layer(
                output,
                memory,
                tgt_mask,
                memory_mask,
                tgt_key_padding_mask,
                memory_key_padding_mask,
                tgt_is_causal,
                memory_is_causal,
            )

        return output


class RotaryEncoderLayer(nn.Module):
    """
    PyTorch API compatible (more or less) encoder layer for Transformers, with support for rotary embeddings.

    Note: Some of the defaults do not match the PyTorch implementation as they are used in my own experiments.

    Args:
        d_model: dimension of the embeddings
        nhead: number of attention heads
        dim_feedforward: dimension of the feedforward network
        dropout: dropout value
        activation: activation function in feedforward (nn.Module)
        norm_first: whether to apply layer norm before or after blocks
        self_rotary: use rotary embeddings in self-attention
    """

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: nn.Module = nn.GELU(),
        norm_first: bool = False,
        self_rotary: bool = True,
    ) -> None:
        super().__init__()

        self.self_attn = RotaryMultiheadAttention(
            d_model, nhead, dropout=dropout, use_rotory_emb=self_rotary
        )

        self._ff_block = FFNSwiGLU(d_model, dim_feedforward, dropout=dropout)

        self.norm_first = norm_first

        self.norm1 = RMSNorm(d_model)
        self.norm2 = RMSNorm(d_model)

        self.activation = activation

        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        src: Tensor,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
    ) -> Tensor:
        x = src

        if self.norm_first:
            x = x + self._sa_block(
                self.norm1(x), src_mask, src_key_padding_mask, is_causal=is_causal
            )
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(
                x
                + self._sa_block(x, src_mask, src_key_padding_mask, is_causal=is_causal)
            )
            x = self.norm2(x + self._ff_block(x))

        return x

    # self-attention block
    def _sa_block(
        self,
        x: Tensor,
        attn_mask: Optional[Tensor],
        key_padding_mask: Optional[Tensor],
        is_causal: bool = False,
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            is_causal=is_causal,
        )
        return self.dropout(x)


class RotaryDecoderLayer(nn.Module):
    """
    PyTorch API compatible (more or less) decoder layer for Transformers, with support for rotary embeddings.

    Note: Some of the defaults do not match the PyTorch implementation, or may be 'strange' (i.e. memory_rotary=False)
    as they are used in my own experiments.

    Args:
        d_model: dimension of the embeddings
        nhead: number of attention heads
        dim_feedforward: dimension of the feedforward network
        dropout: dropout value
        activation: activation function in feedforward (nn.Module)
        norm_first: whether to apply layer norm before or after blocks
        self_rotary: use rotary embeddings in self-attention
        memory_rotary: use rotary embeddings in cross-attention
    """

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: nn.Module = nn.GELU(),
        norm_first: bool = False,
        self_rotary: bool = True,
        memory_rotary: bool = False,
    ):
        super().__init__()

        self.norm_first = norm_first

        self.self_attn = RotaryMultiheadAttention(
            d_model, nhead, dropout=dropout, use_rotory_emb=self_rotary
        )
        self.cross_attn = RotaryMultiheadAttention(
            d_model, nhead, dropout=dropout, use_rotory_emb=memory_rotary
        )

        self.activation = activation

        self._ff_block = FFNSwiGLU(d_model, dim_feedforward, dropout=dropout)

        self.norm1 = RMSNorm(d_model)
        self.norm2 = RMSNorm(d_model)
        self.norm3 = RMSNorm(d_model)

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
    ) -> Tensor:
        x = tgt

        if self.norm_first:
            x = x + self._sa_block(
                self.norm1(x), tgt_mask, tgt_key_padding_mask, is_causal=tgt_is_causal
            )
            x = x + self._mha_block(
                self.norm2(x),
                memory,
                memory_mask,
                memory_key_padding_mask,
                memory_is_causal,
            )
            x = x + self._ff_block(self.norm3(x))
        else:
            x = self.norm1(
                x
                + self._sa_block(
                    x, tgt_mask, tgt_key_padding_mask, is_causal=tgt_is_causal
                )
            )
            x = self.norm2(
                x
                + self._mha_block(
                    x, memory, memory_mask, memory_key_padding_mask, memory_is_causal
                )
            )
            x = self.norm3(x + self._ff_block(x))

        return x

    # self-attention block
    def _sa_block(
        self,
        x: Tensor,
        attn_mask: Optional[Tensor],
        key_padding_mask: Optional[Tensor],
        is_causal: bool = False,
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            is_causal=is_causal,
        )
        return self.dropout1(x)

    # multihead attention block
    def _mha_block(
        self,
        x: Tensor,
        mem: Tensor,
        attn_mask: Optional[Tensor],
        key_padding_mask: Optional[Tensor],
        is_causal: bool = False,
    ) -> Tensor:
        x = self.cross_attn(
            x,
            mem,
            mem,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            is_causal=is_causal,
        )
        return self.dropout2(x)


class RotaryMultiheadAttention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        heads: int,
        dropout: float = 0.0,
        use_rotory_emb: bool = True,
    ):
        super().__init__()

        self.embed_size = embed_dim
        self.num_heads = heads
        self.head_dim = embed_dim // heads
        self.dropout = dropout

        assert (
            self.head_dim * heads == embed_dim
        ), "Embedding size needs to be divisible by heads"

        self.q_proj = nn.Linear(self.embed_size, self.embed_size)
        self.k_proj = nn.Linear(self.embed_size, self.embed_size)
        self.v_proj = nn.Linear(self.embed_size, self.embed_size)
        self.out_proj = nn.Linear(heads * self.head_dim, embed_dim)

        if use_rotory_emb:
            self.rotary_emb = RotaryEmbedding(self.head_dim)

        self.reset_parameters()

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        key_padding_mask: Optional[Tensor] = None,
        attn_mask: Optional[Tensor] = None,
        is_causal: bool = False,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        N = query.shape[0]

        mask = combine_masks(query, key, self.num_heads, attn_mask, key_padding_mask)

        q = self.q_proj(query)
        k = self.k_proj(key)
        v = self.v_proj(value)

        q = q.view(N, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        k = k.view(N, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        v = v.view(N, -1, self.num_heads, self.head_dim).permute(0, 2, 1, 3)

        if hasattr(self, "rotary_emb"):
            q, k = self.rotary_emb(q, k)

        dropout = self.dropout if self.training else 0.0
        attn = F.scaled_dot_product_attention(
            q, k, v, mask, is_causal=is_causal, dropout_p=dropout
        )
        attn = attn.permute(0, 2, 1, 3).reshape(N, -1, self.embed_size)
        return self.out_proj(attn)

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)

        nn.init.constant_(self.q_proj.bias, 0.0)
        nn.init.constant_(self.k_proj.bias, 0.0)
        nn.init.constant_(self.v_proj.bias, 0.0)
        nn.init.constant_(self.out_proj.bias, 0.0)


class SinePositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5_000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-log(10000.0) / d_model))
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[:, : x.shape[1], :]
        return self.dropout(x)


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()

        self.eps = eps
        self.register_parameter("weight", nn.Parameter(torch.ones(dim)))

    def forward(self, x):
        return self.weight.type_as(x) * self._norm(x).type_as(x)

    def _norm(self, x):
        return x * torch.rsqrt(x.square().mean(dim=-1, keepdim=True) + self.eps)


class FFNSwiGLU(nn.Module):
    def __init__(self, dim, dim_feedforward, dropout=0.0):
        super().__init__()

        self.w1 = nn.Linear(dim, dim_feedforward, bias=False)
        self.w2 = nn.Linear(dim, dim_feedforward, bias=False)

        self.w3 = nn.Sequential(
            nn.Dropout(dropout),
            nn.Linear(dim_feedforward, dim, bias=False),
        )

    def forward(self, x):
        swish = F.silu(self.w1(x))
        x = swish * self.w2(x)
        x = self.w3(x)
        return x


def combine_masks(
    q: Tensor,
    k: Tensor,
    heads: int,
    attn_mask: Optional[Tensor],
    key_padding_mask: Optional[Tensor],
) -> Optional[Tensor]:
    """
    Combines the masks for attention and key padding.
    """
    N, L = q.shape[0], q.shape[-2]
    S = k.shape[-2]

    key_padding_mask = F._canonical_mask(
        mask=key_padding_mask,
        mask_name="key_padding_mask",
        other_type=F._none_or_dtype(attn_mask),
        other_name="attn_mask",
        target_type=q.dtype,
    )

    attn_mask = F._canonical_mask(
        mask=attn_mask,
        mask_name="attn_mask",
        other_type=None,
        other_name="",
        target_type=q.dtype,
        check_other=False,
    )

    mask = None
    if attn_mask is not None:
        mask = attn_mask
    if key_padding_mask is not None:
        key_mask = key_padding_mask.view(N, 1, 1, key_padding_mask.shape[-1]).expand(
            N, heads, L, S
        )
        mask = mask + key_mask if mask is not None else key_mask

    return mask


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


def _transformer_reset(module: nn.Module):
    for mod in module.modules():
        if hasattr(mod, "reset_parameters"):
            mod.reset_parameters()
        elif hasattr(mod, "reset_parameters_"):
            mod.reset_parameters_()
        elif hasattr(mod, "_reset_parameters"):
            mod._reset_parameters()

    for mod in module.modules():
        if isinstance(mod, RotaryMultiheadAttention):
            mod.reset_parameters()
        elif isinstance(mod, nn.MultiheadAttention):
            mod._reset_parameters()



# Much of this is inspired by lucidrains ema-pytorch
# https://github.com/lucidrains/ema-pytorch

from enum import Enum, auto
from typing import Any, Dict, List, Optional

import lightning as L
import torch
from lightning import LightningModule, Trainer
from lightning.pytorch.strategies import DDPStrategy, DeepSpeedStrategy, FSDPStrategy
from lightning.pytorch.utilities.exceptions import MisconfigurationException


class UpdateAfter(Enum):
    STEP = auto()
    EPOCH = auto()


class EMACallback(L.Callback):
    """
    Lightning callback to compute a moving average of a models parameters.
    You can easily load the ema

    Attributes:
        tau (float): The decay factor for EMA, should be between 0 and 1. Defaults to 0.9999.
        update_every_n_steps (int): Update the EMA weights every n steps. Defaults to 1 (every step).
        start_on_step (int): Start updating the EMA weights on this step. If None, start immediately.
        start_on_epoch (int): Start updating the EMA weights on this epoch. If None, start immediately.
        excluded_parameters (list of str): List of model parameter names to exclude from EMA.
        use_for_val (bool): If True, use the EMA weights for validation. Defaults to False.
    """

    def __init__(
        self,
        tau: float = 0.9999,
        update_every_n_steps: int = 1,
        start_on_step: Optional[int] = None,
        start_on_epoch: Optional[int] = None,
        excluded_parameters: Optional[List[str]] = None,
        use_for_val: bool = False,
    ):
        super().__init__()

        self.tau = tau

        self.every_n_steps = update_every_n_steps

        if start_on_step is not None and start_on_epoch is not None:
            raise ValueError(
                "Only one of `update_after_n_steps` and `update_after_n_epochs` can be set."
            )

        if start_on_step is not None:
            self.update_start = UpdateAfter.STEP
            self.update_time = start_on_step
        elif start_on_epoch is not None:
            self.update_start = UpdateAfter.EPOCH
            self.update_time = start_on_epoch
        else:
            self.update_start = UpdateAfter.STEP
            self.update_time = 0

        self.excluded_parameters = (
            set(excluded_parameters) if excluded_parameters else set()
        )

        self.ema_weights = None

        # We track the step instead of using the trainer's global_step because
        # multiple optimizers can increment the global step multiple times per batch
        self.step = 0

        # Caching the original state dict for validation
        self._cached_sd = None
        self.use_for_val = use_for_val

        # If we restore from a checkpoint, we need to know if we've transferred the weights
        # to the device yet or not
        self.on_device = False

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs,
        batch,
        batch_idx,
        unused=0,
    ):
        if not self.should_update(trainer):
            return

        self.step += 1

        self._init(trainer, pl_module)

        for name, param in pl_module.named_parameters():
            if name in self.excluded_parameters:
                continue

            # If its not a float just copy it over
            if not param.is_floating_point() and not param.is_complex():
                self.ema_weights[name].copy_(param.data)
                continue

            self.ema_weights[name].lerp_(param.data, 1.0 - self.tau)

    def _init(self, trainer: Trainer, pl_module: LightningModule):
        if self.ema_weights is None:
            self.ema_weights = sd_copy(pl_module.state_dict())
            self.step = 0
            self.on_device = True
        elif not self.on_device:
            self.ema_weights = {
                k: v.to(pl_module.device) for k, v in self.ema_weights.items()
            }
            self.on_device = True

    def should_update(self, trainer: Trainer):
        if self.step % self.every_n_steps != 0:
            return False

        if self.update_start == UpdateAfter.STEP:
            return self.step >= self.update_time
        elif self.update_start == UpdateAfter.EPOCH:
            return trainer.current_epoch >= self.update_time

    @staticmethod
    def load_ema_weights(pl_module: LightningModule, ckpt_path: str, **kwargs) -> None:
        """Load the EMA weights from a checkpoint"""

        sd = torch.load(ckpt_path)
        ema_weights = sd["callbacks"]["EMACallback"]["state_dict"]
        pl_module.load_state_dict(ema_weights, **kwargs)

    def on_validation_start(self, trainer: Trainer, pl_module: LightningModule):
        """Swap the weights for validation"""
        if self.use_for_val and self.ema_weights is not None:
            self._cached_sd = sd_copy(pl_module.state_dict())
            pl_module.load_state_dict(self.ema_weights)

    def on_validation_end(self, trainer: Trainer, pl_module: LightningModule):
        """Swap the weights back after validation"""
        if self.use_for_val and self._cached_sd is not None:
            pl_module.load_state_dict(self._cached_sd)
            self._cached_sd = None

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if isinstance(trainer.strategy, (FSDPStrategy, DeepSpeedStrategy, DDPStrategy)):
            raise MisconfigurationException(
                "I haven't tested this with FSDP, DeepSpeed, or DDP. Don't use it."
            )

    def state_dict(self) -> Dict[str, Any]:
        return {
            "state_dict": self.ema_weights,
            "step": self.step,
            "update_start": self.update_start,
            "update_time": self.update_time,
            "excluded_parameters": self.excluded_parameters,
            "use_for_val": self.use_for_val,
            "tau": self.tau,
            "every_n_steps": self.every_n_steps,
        }

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        self.ema_weights = state_dict["state_dict"]
        self.step = state_dict["step"]
        self.update_start = state_dict["update_start"]
        self.update_time = state_dict["update_time"]
        self.excluded_parameters = state_dict["excluded_parameters"]
        self.use_for_val = state_dict["use_for_val"]
        self.tau = state_dict["tau"]
        self.every_n_steps = state_dict["every_n_steps"]


def sd_copy(state_dict):
    """Probably faster than deepcopy idk"""
    return {k: v.clone().detach() for k, v in state_dict.items()}



import torch
from torch import Tensor
import math


def upbounded_omega(p: Tensor):
    """
    More stable parameterization of variance, in addition the variance is capped at 1.
    See https://arxiv.org/abs/2106.13739 eqn. 12 for details
    """
    mask = p <= 0

    pexp = p.exp()
    npexp = (-p).exp()

    sigma = torch.where(
        mask,
        0.5 * pexp,
        0.5 * (2.0 - npexp),
    )

    lo = math.log(0.5)
    # Clamp to avoid log(0), 1e-8 is not representable in fp16 (it is in bfloat16 though), so we use 1e-5 to be safe
    # Even though the NaNs/nonfinites are masked out, they still can cause issues with gradient propagation
    logsigma = torch.where(
        mask,
        p + lo,
        (2.0 - npexp).clamp_min(1e-5).log() + lo,
    )

    return sigma.type_as(p), logsigma.type_as(p)


def gaussian_kl(
    mu1: Tensor,
    logsigma1: Tensor,
    mu2: Tensor,
    logsigma2: Tensor,
):
    """
    More numerically stable diagonal Gaussian KL (dont want to exp then log)
    Make sure to pass in log(sigma), not sigma
    """
    return (
        0.5
        * ((2.0 * logsigma1).exp() + (mu1 - mu2).square())
        * (-2.0 * logsigma2).exp()
        + logsigma2
        - logsigma1
        - 0.5
    )



from hnn_utils.callbacks.ema import EMACallback
from hnn_utils.nn.Transformer import (
    RotaryDecoder,
    RotaryDecoderLayer,
    RotaryEncoder,
    RotaryEncoderLayer,
    SinePositionalEncoding,
)



from setuptools import setup, find_packages

setup(
    name="hnn_utils",
    version="0.1.18",
    packages=find_packages(),
    install_requires=[
        "torch",
        "lightning",
    ],
    author="Haydn Jones",
    author_email="haydnjonest@gmail.com",
    description="Various utilities used throughout my research",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/haydn-jones/hnn_utils",
)



from hnn_utils.nn.RotaryEmbedding import RotaryEmbedding
from hnn_utils.nn.Transformer import (
    RotaryDecoder,
    RotaryDecoderLayer,
    RotaryEncoder,
    RotaryEncoderLayer,
    RotaryMultiheadAttention,
)
